
/******************************************************************************/
/*  (c)	Copyright Hewlett-Packard Company, 1993,  Fort Collins,	Colorado      */
/*									      */
/*  (C)	COPYRIGHT International	Business Machines Corp.	1993		      */
/*									      */
/*   Copyright (c) 1993	by Sun Microsystems, Inc.			      */
/*									      */
/*			      All Rights Reserved			      */
/*									      */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for any purpose and without fee is hereby granted,	      */
/*  provided that the above copyright notices appear in	all copies and that   */
/*  both the copyright notices and this	permission notice appear in	      */
/*  supporting documentation, and that the names of Hewlett-Packard,	      */
/*  International Business Machines, or	Sun Microsystems not be	used in	      */
/*  advertising	or publicity pertaining	to distribution	of the software	      */
/*  without specific, written prior permission.				      */
/*									      */
/*  HEWLETT-PACKARD MAKES NO WARRANTY OF ANY KIND WITH REGARD TO THIS	      */
/*  SOFTWARE, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF	      */
/*  MERCHANTABILITY AND	FITNESS	FOR A PARTICULAR PURPOSE.  Hewlett-Packard    */
/*  shall not be liable	for errors contained herein or direct, indirect,      */
/*  special, incidental	or consequential damages in connection with the	      */
/*  furnishing,	performance or use of this software.			      */
/*									      */
/*  INTERNATIONAL BUSINESS MACHINES CORPORATION	PROVIDES THIS MATERIAL	      */
/*  "AS	IS" WITHOUT WARRANTY OF	ANY KIND, EITHER EXPRESSED OR IMPLIED,	      */
/*  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY  */
/*  AND	FITNESS	FOR A PARTICULAR PURPOSE.				      */
/*									      */
/*  SUN	MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES WITH REGARD	TO THIS	      */
/*  SOFTWARE, INCLUDING	ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,*/
/*  IN NO EVENT	SHALL SUN MICROSYSTEMS,	INC. BE	LIABLE FOR ANY SPECIAL,	      */
/*  INDIRECT OR	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER	RESULTING     */
/*  FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,      */
/*  NEGLIGENCE OR OTHER	TORTIOUS ACTION, ARISING OUT OF	OR IN CONNECTION      */
/*  WITH THE USE OR PERFORMANCE	OF THIS	SOFTWARE.			      */
/*									      */
/******************************************************************************/

/*
 * This is a SAMPLE program for illustration purposes only.
 * IBM makes no commitment for support of this module.
 */

/*
===============================================================================
	Common Graphics	Environment (CGE) extensions
	PEX 5.1	Simple Programming Example

	View Computations

===============================================================================
*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <X11/PEX5/PEXlib.h>

#include "ptorus.h"


/*
*******************************************************************************
Linear algebra utilities.
*******************************************************************************
*/

void vector_subtract(c,	a, b)
float *c;
float *a;
float *b;
{
    *c++ = *a++	- *b++;
    *c++ = *a++	- *b++;
    *c++ = *a++	- *b++;
}

double vector_length(v)
    register float *v;
{
    register int i;
    register double s;

    for	(i = 0,	s = 0.0; i < 3;	i++, v++)
	s += *v	* *v;
    s =	sqrt(s);
    return(s);
}

void vector_scale(v, s)
    register float *v, s;
{
    *v++ *= s;
    *v++ *= s;
    *v++ *= s;
}


/*
*******************************************************************************
    This procedure creates a view table	entry and installs it.
*******************************************************************************
*/

void set_view_camera(dpy, window, ca, renderer,	rend_attr, index)
    Display			*dpy;
    Window			window;
    camera_arg			*ca;
    PEXRenderer			renderer;
    PEXRendererAttributes	*rend_attr;
    int				index;
{
    double viewDistance;

    PEXCoord		viewReferencePoint;
    PEXCoord		projectionReferencePoint;
    PEXVector		viewUpVector, viewPlaneNormal;
    PEXCoord2D		viewWindow[2];
    double		viewPlane, frontPlane, backPlane;
    PEXNPCSubVolume	viewPort;
    PEXViewEntry	view;
    int			err, perspective;
    double		windowScale;
    double		winAspect;
    int			winWidth, winHeight;

    XWindowAttributes	win_attrs;

    /*
     * Compute the ViewOrientationMatrix from viewReferencePoint,
     * viewPlaneNormal and viewUpVector.  This stuff is	independent of the
     * resize method.
     */

    viewReferencePoint.x = ca->refx;
    viewReferencePoint.y = ca->refy;
    viewReferencePoint.z = ca->refz;

    vector_subtract((float *) &viewPlaneNormal,	&(ca->camx), &(ca->refx));
    viewDistance = vector_length(&viewPlaneNormal);
    vector_scale(&viewPlaneNormal, 1.0 / viewDistance);

    viewPlane  = 0;
    backPlane  = -10000;
    frontPlane = viewDistance -	1;

    viewUpVector.x = ca->upx;
    viewUpVector.y = ca->upy;
    viewUpVector.z = ca->upz;

    err	= PEXViewOrientationMatrix(&viewReferencePoint,
				   &viewPlaneNormal,
				   &viewUpVector,
				   view.orientation);

    /*
       Compute the size	(and shape) of the viewWindow based upon the camera's
       field of	view and the aspect ratio of the window	to which we're about
       to render.
    */

    XGetWindowAttributes (dpy, window, &win_attrs);

    windowScale	= atan(ca->field_of_view * M_PI	/ 360.0)
			* (viewDistance	- viewPlane);
    winWidth	= (win_attrs.width  > 1) ? (win_attrs.width  - 1) : 1;
    winHeight	= (win_attrs.height > 1) ? (win_attrs.height - 1) : 1;
    winAspect	= (double) winWidth / (double) winHeight;

    viewPort.min.x = 0;	   viewPort.max.x = 1;
    viewPort.min.y = 0;	   viewPort.max.y = 1;
    viewPort.min.z = 0;	   viewPort.max.z = 1;
    rend_attr->npc_subvolume = viewPort;

    viewWindow[0].y = -windowScale;
    viewWindow[1].y =  windowScale;
    viewWindow[0].x = -windowScale;
    viewWindow[1].x =  windowScale;

    view.clip_flags = PEXClippingAll;
    perspective	= True;
    projectionReferencePoint.x = 0;
    projectionReferencePoint.y = 0;
    projectionReferencePoint.z = viewDistance;

    /*
     * Compute the aspect ratio	of the viewPort	so we fill the window.	From
     * that the	ViewMappingMatrix is computed.
     */


    /*
     * This method varies only the renderer's npc_subvolume to map the 0-1
     * range to	the major dimension of the destination drawable.  The minor
     * dimension is reduced by the aspect ratio	of the drawable	and then
     * centered.
     */

    if (winWidth > winHeight)
    {
	double aspect =	(double) winHeight / (double) winWidth;
	
	rend_attr->npc_subvolume.min.y = (1 - aspect) /	2;
	rend_attr->npc_subvolume.max.y =
				rend_attr->npc_subvolume.min.y + aspect;
    }
    else if (winWidth <	winHeight)
    {
	rend_attr->npc_subvolume.min.x = (1 - winAspect) / 2;
	rend_attr->npc_subvolume.max.x =
				rend_attr->npc_subvolume.min.x + winAspect;
    }

    view.clip_limits = viewPort;
    PEXChangeRenderer(dpy, renderer, PEXRANPCSubVolume,	rend_attr);

    err	= PEXViewMappingMatrix(viewWindow, &viewPort, perspective,
			       &projectionReferencePoint,
			       viewPlane, backPlane, frontPlane,
			       view.mapping);

    PEXSetTableEntries(dpy, rend_attr->view_table, index, 1, PEXLUTView,
			(PEXPointer) &view);

} /* set_view_camera */
